@@ -58,6 +58,54 @@ static sql_fbh_t ora2sql_type _((imp_fbh_t* fbh));
 void ora_free_phs_contents _((phs_t *phs));
 static void dump_env_to_trace();
 
+static sb4
+oci_error_get(OCIError *errhp, sword status, char *what, SV *errstr, int debug)
+{
+    text errbuf[1024];
+    ub4 recno = 0;
+    sb4 errcode = 0;
+    sb4 eg_errcode = 0;
+    sword eg_status;
+
+    if (!SvOK(errstr))
+	sv_setpv(errstr,"");
+    if (!errhp) {
+	sv_catpv(errstr, oci_status_name(status));
+	if (what) {
+	    sv_catpv(errstr, " ");
+	    sv_catpv(errstr, what);
+	}
+	return status;
+    }
+
+    while( ++recno
+	&& OCIErrorGet_log_stat(errhp, recno, (text*)NULL, &eg_errcode, errbuf,
+	    (ub4)sizeof(errbuf), OCI_HTYPE_ERROR, eg_status) != OCI_NO_DATA
+	&& eg_status != OCI_INVALID_HANDLE
+	&& recno < 100
+    ) {
+	if (debug >= 4 || recno>1/*XXX temp*/)
+	    PerlIO_printf(DBILOGFP, "    OCIErrorGet after %s (er%ld:%s): %d, %ld: %s\n",
+		what ? what : "<NULL>", (long)recno,
+		    (eg_status==OCI_SUCCESS) ? "ok" : oci_status_name(eg_status),
+		    status, (long)eg_errcode, errbuf);
+	errcode = eg_errcode;
+	sv_catpv(errstr, (char*)errbuf);
+	if (*(SvEND(errstr)-1) == '\n')
+	    --SvCUR(errstr);
+    }
+    if (what || status != OCI_ERROR) {
+	sv_catpv(errstr, (debug<0) ? " (" : " (DBD ");
+	sv_catpv(errstr, oci_status_name(status));
+	if (what) {
+	    sv_catpv(errstr, ": ");
+	    sv_catpv(errstr, what);
+	}
+	sv_catpv(errstr, ")");
+    }
+    return errcode;
+}
+
 static int
 GetRegKey(char *key, char *val, char *data, unsigned long *size)
 {
@@ -104,7 +152,7 @@ ora_env_var(char *name, char *buf, unsigned long size)
 /* Under Cygwin there are issues with setting environment variables
  * at runtime such that Windows-native libraries loaded by a Cygwin
  * process can see those changes.
- * 
+ *
  * Cygwin maintains its own cache of environment variables, and also
  * only writes to the Windows environment using the "_putenv" win32
  * call. This call writes to a Windows C runtime cache, rather than
@@ -113,7 +161,7 @@ ora_env_var(char *name, char *buf, unsigned long size)
  * In order to change environment variables so that the Oracle client
  * DLL can see the change, the win32 function SetEnvironmentVariable
  * must be called. This function gives an interface to that API.
- * 
+ *
  * It is only available when building under Cygwin, and is used by
  * the testsuite.
  *
@@ -255,7 +303,7 @@ typedef struct {
     /*recursive_lock_t    lock; */
     /*perl_cond           user_cond;*/      /* For user-level conditions */
 } shared_sv;
-	
+
 
 
 int
@@ -282,26 +330,26 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
     shared_dbh_priv_svp = (DBD_ATTRIB_OK(attr)?hv_fetch((HV*)SvRV(attr), "ora_dbh_share", 13, 0):NULL) ;
     shared_dbh_priv_sv = shared_dbh_priv_svp?*shared_dbh_priv_svp:NULL ;
 
-    if (shared_dbh_priv_sv && SvROK(shared_dbh_priv_sv)) 
-	shared_dbh_priv_sv = SvRV(shared_dbh_priv_sv) ;	
-    
+    if (shared_dbh_priv_sv && SvROK(shared_dbh_priv_sv))
+	shared_dbh_priv_sv = SvRV(shared_dbh_priv_sv) ;
+
     if (shared_dbh_priv_sv) {
 	MAGIC * mg ;
 
 	SvLOCK (shared_dbh_priv_sv) ;
-	
+
         /* some magic from shared.xs (no public api yet :-( */
 	mg = mg_find(shared_dbh_priv_sv, PERL_MAGIC_shared_scalar) ;
-	
+
 	shared_dbh_ssv = (shared_sv * )(mg?mg -> mg_ptr:NULL) ;  /*sharedsv_find(*shared_dbh_priv_sv) ;*/
 	if (!shared_dbh_ssv)
 	    croak ("value of ora_dbh_share must be a scalar that is shared") ;
-		
+
 	shared_dbh 		= (imp_dbh_t *)SvPVX(shared_dbh_ssv -> sv) ;
 	shared_dbh_len 	= SvCUR((shared_dbh_ssv -> sv)) ;
-	if (shared_dbh_len > 0 && shared_dbh_len != sizeof (imp_dbh_t)) 
+	if (shared_dbh_len > 0 && shared_dbh_len != sizeof (imp_dbh_t))
 	    croak ("Invalid value for ora_dbh_dup") ;
-		
+
 	if (shared_dbh_len == sizeof (imp_dbh_t)) {
 	    /* initialize from shared data */
             memcpy (((char *)imp_dbh) + DBH_DUP_OFF, ((char *)shared_dbh) + DBH_DUP_OFF, DBH_DUP_LEN) ;
@@ -314,7 +362,7 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
             shared_dbh = NULL ;
        }
     }
-#endif	
+#endif
 
     /* Check if we should re-use a ProC connection and not connect ourselves. */
     DBD_ATTRIB_GET_IV(attr, "ora_use_proc_connection", 23,
@@ -453,9 +501,9 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
                     "OCIEnvNlsCreate. Check ORACLE_HOME env var, NLS settings, permissions, etc.");
                 return 0;
             }
-                
+
             /* update the hard-coded csid constants for unicode charsets */
-            utf8_csid      = OCINlsCharSetNameToId(imp_dbh->envhp, (void*)"UTF8"); 
+            utf8_csid      = OCINlsCharSetNameToId(imp_dbh->envhp, (void*)"UTF8");
             al32utf8_csid  = OCINlsCharSetNameToId(imp_dbh->envhp, (void*)"AL32UTF8");
             al16utf16_csid = OCINlsCharSetNameToId(imp_dbh->envhp, (void*)"AL16UTF16");
 
@@ -531,7 +579,7 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
 	/* We don't have a way to get the actual charsetid & ncharsetid in use
 	*  but we only care about UTF8 so we'll just check for that and use the
 	*  the hardcoded utf8_csid if found
-	*/  
+	*/
 	char buf[81];
 	char *nls = ora_env_var("NLS_LANG", buf, sizeof(buf)-1);
 	if (nls && strlen(nls) >= 4 && !strcasecmp(nls + strlen(nls) - 4, "utf8"))
@@ -540,10 +588,10 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
 	if (nls && strlen(nls) >= 4 && !strcasecmp(nls + strlen(nls) - 4, "utf8"))
 	     ncharsetid = utf8_csid;
     }
-#endif 
+#endif
 #endif
 
-    /* At this point we have charsetid & ncharsetid 
+    /* At this point we have charsetid & ncharsetid
     *  note that it is possible for charsetid and ncharestid to
     *  be distinct if NLS_LANG and NLS_NCHAR are both used.
     *  BTW: NLS_NCHAR is set as follows: NSL_LANG=AL32UTF8
@@ -603,7 +651,7 @@ dbd_db_login6(SV *dbh, imp_dbh_t *imp_dbh, char *dbname, char *uid, char *pwd, S
 		return 0;
 	    }
 
-	    OCIAttrSet_log_stat( imp_dbh->svchp, OCI_HTYPE_SVCCTX, imp_dbh->srvhp, 
+	    OCIAttrSet_log_stat( imp_dbh->svchp, OCI_HTYPE_SVCCTX, imp_dbh->srvhp,
 			    (ub4) 0, OCI_ATTR_SERVER, imp_dbh->errhp, status);
 
 	    OCIHandleAlloc_ok(imp_dbh->envhp, &imp_dbh->authp, OCI_HTYPE_SESSION, status);
@@ -657,7 +705,7 @@ dbd_db_login6_out:
 	memcpy(SvPVX(shared_dbh_priv_sv) + DBH_DUP_OFF, ((char *)imp_dbh) + DBH_DUP_OFF, DBH_DUP_LEN) ;
 	SvSETMAGIC(shared_dbh_priv_sv);
 	imp_dbh->shared_dbh = (imp_dbh_t *)SvPVX(shared_dbh_ssv->sv);
-    }		
+    }
 #endif
 
     return 1;
@@ -746,10 +794,10 @@ dbd_db_disconnect(SV *dbh, imp_dbh_t *imp_dbh)
 void
 dbd_db_destroy(SV *dbh, imp_dbh_t *imp_dbh)
 {
-    dTHX ;	
+    dTHX ;
     int refcnt = 1 ;
     sword status;
-	
+
 #if defined(USE_ITHREADS) && defined(PERL_MAGIC_shared_scalar)
     if (DBIc_IMPSET(imp_dbh) && imp_dbh->shared_dbh) {
 	SvLOCK (imp_dbh->shared_dbh_priv_sv) ;
@@ -873,6 +921,7 @@ dbd_preparse(imp_sth_t *imp_sth, char *statement)
     int idx=0;
     char *style="", *laststyle=Nullch;
     STRLEN namelen;
+    phs_t *phs;
 
     /* allocate room for copy of statement with spare capacity	*/
     /* for editing '?' or ':1' into ':p1' so we can use obndrv.	*/
@@ -941,7 +990,7 @@ dbd_preparse(imp_sth_t *imp_sth, char *statement)
 
 	/* only here for : or ? outside of a comment or literal	*/
 
-	start = dest;			/* save name inc colon	*/ 
+	start = dest;			/* save name inc colon	*/
 	*dest++ = *src++;
 	if (*start == '?') {		/* X/Open standard	*/
 	    sprintf(start,":p%d", ++idx); /* '?' -> ':p1' (etc)	*/
@@ -973,8 +1022,11 @@ dbd_preparse(imp_sth_t *imp_sth, char *statement)
 	if (imp_sth->all_params_hv == NULL)
 	    imp_sth->all_params_hv = newHV();
 	phs_sv = newSVpv((char*)&phs_tpl, sizeof(phs_tpl)+namelen+1);
+	phs = (phs_t*)(void*)SvPVX(phs_sv);
 	hv_store(imp_sth->all_params_hv, start, namelen, phs_sv, 0);
-	strcpy( ((phs_t*)(void*)SvPVX(phs_sv))->name, start);
+	phs->idx = idx-1;       /* Will be 0 for :1, -1 for :foo. */
+    strcpy(phs->name, start);
+
     }
     *dest = '\0';
     if (imp_sth->all_params_hv) {
@@ -1036,7 +1088,7 @@ ora_sql_type(imp_sth_t *imp_sth, char *name, int sql_type)
 
 
 
-static int 
+static int
 dbd_rebind_ph_char(SV *sth, imp_sth_t *imp_sth, phs_t *phs, ub2 **alen_ptr_ptr)
 {
     STRLEN value_len;
@@ -1055,7 +1107,7 @@ dbd_rebind_ph_char(SV *sth, imp_sth_t *imp_sth, phs_t *phs, ub2 **alen_ptr_ptr)
     if (DBIS->debug >= 2) {
 	char *val = neatsvpv(phs->sv,0);
  	PerlIO_printf(DBILOGFP, "       bind %s <== %.1000s (", phs->name, val);
- 	if (!SvOK(phs->sv)) 
+ 	if (!SvOK(phs->sv))
 	    PerlIO_printf(DBILOGFP, "NULL, ");
 	PerlIO_printf(DBILOGFP, "size %ld/%ld/%ld, ",
 	    (long)SvCUR(phs->sv),(long)SvLEN(phs->sv),phs->maxlen);
@@ -1081,7 +1133,7 @@ dbd_rebind_ph_char(SV *sth, imp_sth_t *imp_sth, phs_t *phs, ub2 **alen_ptr_ptr)
 	    /* phs->sv _is_ the real live variable, it may 'mutate' later	*/
 	    /* pre-upgrade to high'ish type to reduce risk of SvPVX realloc/move */
 	    (void)SvUPGRADE(phs->sv, SVt_PVNV);
-	    SvGROW(phs->sv, (STRLEN)((phs->maxlen < min_len) ? min_len : phs->maxlen)+1/*for null*/);
+	    SvGROW(phs->sv, (STRLEN)(((unsigned int) phs->maxlen < min_len) ? min_len : (unsigned int) phs->maxlen)+1/*for null*/);
 	}
     }
 
@@ -1125,8 +1177,8 @@ dbd_rebind_ph_char(SV *sth, imp_sth_t *imp_sth, phs_t *phs, ub2 **alen_ptr_ptr)
  * This allows passing cursor refs as "in" to pl/sql (but only if you got the
  * cursor from pl/sql to begin with)
  */
-int 
-pp_rebind_ph_rset_in(SV *sth, imp_sth_t *imp_sth, phs_t *phs) 
+int
+pp_rebind_ph_rset_in(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 {
     /*dTHR; -- do we need to do this??? */
     SV * sth_csr = phs->sv;
@@ -1158,7 +1210,7 @@ pp_rebind_ph_rset_in(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 
 
 int
-pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec) 
+pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec)
 {
     if (pre_exec) {	/* pre-execute - allocate a statement handle */
 	dSP;
@@ -1248,7 +1300,7 @@ pp_exec_rset(SV *sth, imp_sth_t *imp_sth, phs_t *phs, int pre_exec)
 }
 
 
-static int 
+static int
 dbd_rebind_ph(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 {
     ub2 *alen_ptr = NULL;
@@ -1332,10 +1384,10 @@ dbd_rebind_ph(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 
     if (csform) {
     	/* set OCI_ATTR_CHARSET_FORM before we get the default OCI_ATTR_CHARSET_ID */
-	OCIAttrSet_log_stat(phs->bndhp, (ub4) OCI_HTYPE_BIND, 
+	OCIAttrSet_log_stat(phs->bndhp, (ub4) OCI_HTYPE_BIND,
 	    &csform, (ub4) 0, (ub4) OCI_ATTR_CHARSET_FORM, imp_sth->errhp, status);
 	if ( status != OCI_SUCCESS ) {
-	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_CHARSET_FORM)")); 
+	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_CHARSET_FORM)"));
 	    return 0;
 	}
     }
@@ -1364,10 +1416,10 @@ dbd_rebind_ph(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 
 
     if (csid) {
-	OCIAttrSet_log_stat(phs->bndhp, (ub4) OCI_HTYPE_BIND, 
+	OCIAttrSet_log_stat(phs->bndhp, (ub4) OCI_HTYPE_BIND,
 	    &csid, (ub4) 0, (ub4) OCI_ATTR_CHARSET_ID, imp_sth->errhp, status);
 	if ( status != OCI_SUCCESS ) {
-	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_CHARSET_ID)")); 
+	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_CHARSET_ID)"));
 	    return 0;
 	}
     }
@@ -1376,7 +1428,7 @@ dbd_rebind_ph(SV *sth, imp_sth_t *imp_sth, phs_t *phs)
 	OCIAttrSet_log_stat(phs->bndhp, (ub4)OCI_HTYPE_BIND,
 	    neatsvpv(phs->sv,0), (ub4)phs->maxdata_size, (ub4)OCI_ATTR_MAXDATA_SIZE, imp_sth->errhp, status);
 	if ( status != OCI_SUCCESS ) {
-	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_MAXDATA_SIZE)")); 
+	    oci_error(sth, imp_sth->errhp, status, ora_sql_error(imp_sth,"OCIAttrSet (OCI_ATTR_MAXDATA_SIZE)"));
 	    return 0;
 	}
     }
@@ -1417,6 +1469,7 @@ dbd_bind_ph(SV *sth, imp_sth_t *imp_sth, SV *ph_namesv, SV *newvalue, IV sql_typ
     if (SvROK(newvalue)
 	&& !IS_DBI_HANDLE(newvalue)	/* dbi handle allowed for cursor variables */
 	&& !SvAMAGIC(newvalue)		/* overload magic allowed (untested) */
+    && !sv_derived_from(newvalue, "OCILobLocatorPtr" )  /* input LOB locator*/
     )
 	croak("Can't bind a reference (%s)", neatsvpv(newvalue,0));
     if (SvTYPE(newvalue) > SVt_PVLV) /* hook for later array logic?	*/
@@ -1729,7 +1782,7 @@ dbd_st_execute(SV *sth, imp_sth_t *imp_sth) /* <= -2:error, >=0:ok row count, (-
 		AV *av = (AV*)SvRV(sv);
 		I32 avlen = AvFILL(av);
 		if (avlen >= 0)
-		    dbd_phs_avsv_complete(phs, avlen, debug); 
+		    dbd_phs_avsv_complete(phs, avlen, debug);
 	    }
 	    else
 		dbd_phs_sv_complete(phs, sv, debug);
@@ -1739,6 +1792,356 @@ dbd_st_execute(SV *sth, imp_sth_t *imp_sth) /* <= -2:error, >=0:ok row count, (-
     return row_count;	/* row count (0 will be returned as "0E0")	*/
 }
 
+static int
+do_bind_array_exec(sth, imp_sth, phs)
+    SV *sth;
+    imp_sth_t *imp_sth;
+    phs_t *phs;
+{
+    sword status;
+
+    OCIBindByName_log_stat(imp_sth->stmhp, &phs->bndhp, imp_sth->errhp,
+            (text*)phs->name, (sb4)strlen(phs->name),
+            0,
+            phs->maxlen ? (sb4)phs->maxlen : 1, /* else bind "" fails */
+            (ub2)phs->ftype, 0,
+            NULL, /* ub2 *alen_ptr not needed with OCIBindDynamic */
+            0,
+            0,      /* max elements that can fit in allocated array */
+            NULL, /* (ptr to) current number of elements in array */
+            (ub4)OCI_DATA_AT_EXEC,
+            status);
+    if (status != OCI_SUCCESS) {
+        oci_error(sth, imp_sth->errhp, status, "OCIBindByName");
+        return 0;
+    }
+    OCIBindDynamic_log(phs->bndhp, imp_sth->errhp,
+                       (dvoid *)phs, dbd_phs_in,
+                       (dvoid *)phs, dbd_phs_out, status);
+    if (status != OCI_SUCCESS) {
+        oci_error(sth, imp_sth->errhp, status, "OCIBindDynamic");
+        return 0;
+    }
+    return 1;
+}
+
+static void
+init_bind_for_array_exec(phs)
+    phs_t *phs;
+{
+    if (phs->sv == &sv_undef) { /* first bind for this placeholder  */
+        phs->is_inout = 0;
+        phs->maxlen = 1;
+        /* treat Oracle7 SQLT_CUR as SQLT_RSET for Oracle8 */
+        if (phs->ftype==102)
+            phs->ftype = 116;
+        /* some types require the trailing null included in the length. */
+        /* SQLT_STR=5=STRING, SQLT_AVC=97=VARCHAR */
+        phs->alen_incnull = (phs->ftype==SQLT_STR || phs->ftype==SQLT_AVC);
+    }
+}
+
+/* Re-bind placeholders for array execute, to get the correct
+   max-length values. */
+static int
+ora_st_bind_for_array_exec(sth, imp_sth, tuples_av, exe_count, param_count, columns_av)
+    SV *sth;
+    imp_sth_t *imp_sth;
+    AV *tuples_av;
+    ub4 exe_count;
+    int param_count;
+    AV *columns_av;
+{
+    int i, j;
+    char namebuf[30];
+    SV **sv_p;
+    SV *sv;
+    AV *av;
+    phs_t **phs;
+    STRLEN len;
+    /*sword status;*/
+
+    phs = safemalloc(param_count*sizeof(*phs));
+    memset(phs, 0, param_count*sizeof(*phs));
+
+    /* Loop over values, computing maximum lengths. */
+    if(columns_av) {
+        /* Column-wise operation; tuples_av holds a list of columns. */
+        for(i = 0; i < param_count; i++) {
+            char *name;
+
+            sv_p = av_fetch(columns_av, i, 0);
+            if(sv_p == NULL) {
+                Safefree(phs);
+                croak("Missing column entry %d", i);
+            }
+            sv = *sv_p;
+            if (DBIS->debug >= 9)
+                PerlIO_printf(DBILOGFP, "   arrbind %d->'%s'\n", i, neatsvpv(sv,0));
+            name = SvPV(sv, len);
+            sv_p = hv_fetch(imp_sth->all_params_hv, name, len, 0);
+            if (sv_p == NULL) {
+                Safefree(phs);
+                croak("Can't execute for non-existent placeholder %s",
+                      neatsvpv(sv,0));
+            }
+            phs[i] = (phs_t*)(void*)SvPVX(*sv_p); /* placeholder struct */
+            phs[i]->idx = i;
+            init_bind_for_array_exec(phs[i]);
+
+            sv_p = av_fetch(tuples_av, i, 0);
+            if(sv_p == NULL) {
+                Safefree(phs);
+                croak("Cannot fetch column %d from tuple array", i);
+            }
+            sv = *sv_p;
+            if(!SvROK(sv) || SvTYPE(SvRV(sv)) != SVt_PVAV) {
+                Safefree(phs);
+                croak("Not an array ref in column %d", i);
+            }
+            av = (AV*)SvRV(sv);
+
+            for(j = 0; (unsigned int) j <  exe_count; j++) {
+                sv_p = av_fetch(av, j, 0);
+                if(sv_p == NULL) {
+                    Safefree(phs);
+                    croak("Cannot fetch value for param %d in element %d", i, j);
+                }
+                sv = *sv_p;
+
+                /* Find the value length, and increase maxlen if needed. */
+                if(SvROK(sv)) {
+                    Safefree(phs);
+                    croak("Can't bind a reference (%s) for param %d, entry %d",
+                          neatsvpv(sv,0), i, j);
+                }
+                SvPV(sv, len);
+                if(len > (unsigned int) phs[i]->maxlen)
+                    phs[i]->maxlen = len;
+            }
+
+            if(!do_bind_array_exec(sth, imp_sth, phs[i])) {
+                Safefree(phs);
+                return 0;
+            }
+        }
+    } else {
+        /* Row-wise operation; tuples_av holds a list of bind value tuples. */
+        for(j = 0; (unsigned int) j < exe_count; j++) {
+            sv_p = av_fetch(tuples_av, j, 0);
+            if(sv_p == NULL) {
+                Safefree(phs);
+                croak("Cannot fetch tuple %d", j);
+            }
+            sv = *sv_p;
+            if(!SvROK(sv) || SvTYPE(SvRV(sv)) != SVt_PVAV) {
+                Safefree(phs);
+                croak("Not an array ref in element %d", j);
+            }
+            av = (AV*)SvRV(sv);
+            for(i = 0; i < param_count; i++) {
+                if(!phs[i]) {
+                    SV **phs_svp;
+
+                    sprintf(namebuf, ":p%d", i+1);
+                    phs_svp = hv_fetch(imp_sth->all_params_hv,
+                                       namebuf, strlen(namebuf), 0);
+                    if (phs_svp == NULL) {
+                        Safefree(phs);
+                        croak("Can't execute for non-existent placeholder :%d", i);
+                    }
+                    phs[i] = (phs_t*)(void*)SvPVX(*phs_svp); /* placeholder struct */
+                    if(phs[i]->idx < 0) {
+                        Safefree(phs);
+                        croak("Placeholder %d not of ?/:1 type", i);
+                    }
+                    init_bind_for_array_exec(phs[i]);
+                }
+
+                sv_p = av_fetch(av, phs[i]->idx, 0);
+                if(sv_p == NULL) {
+                    Safefree(phs);
+                    croak("Cannot fetch value for param %d in entry %d", i, j);
+                }
+                sv = *sv_p;
+
+                /* Find the value length, and increase maxlen if needed. */
+                if(SvROK(sv)) {
+                    Safefree(phs);
+                    croak("Can't bind a reference (%s) for param %d, entry %d",
+                          neatsvpv(sv,0), i, j);
+                }
+                SvPV(sv, len);
+                if(len > (unsigned int) phs[i]->maxlen)
+                    phs[i]->maxlen = len;
+
+                /* Do OCI bind calls on last iteration. */
+                if(j == exe_count - 1) {
+                  if(!do_bind_array_exec(sth, imp_sth, phs[i])) {
+                    Safefree(phs);
+                    return 0;
+                  }
+                }
+            }
+            /* ToDo: Maybe extract common code from here and dbd_bind_ph() and put
+               it in some function(s) that can be called from both places. */
+        }
+    }
+
+    Safefree(phs);
+    return 1;                   /* Success. */
+}
+
+ int
+ora_st_execute_array(sth, imp_sth, tuples, tuples_status, columns, exe_count)
+    SV *sth;
+    imp_sth_t *imp_sth;
+    SV *tuples;
+    SV *tuples_status;
+    SV *columns;
+    ub4 exe_count;
+{
+    dTHR;
+    /*ub4 row_count = 0;*/
+    int debug = DBIS->debug;
+    D_imp_dbh_from_sth;
+    sword status, exe_status;
+    int is_select = (imp_sth->stmt_type == OCI_STMT_SELECT);
+    AV *tuples_av, *tuples_status_av, *columns_av;
+    ub4 oci_mode;
+    ub4 num_errs;
+    int i;
+    int autocommit = DBIc_has(imp_dbh,DBIcf_AutoCommit);
+
+    if (debug >= 2)
+ PerlIO_printf(DBILOGFP, "    ora_st_execute_array %s count=%d (%s %s %s)...\n",
+                      oci_stmt_type_name(imp_sth->stmt_type), exe_count,
+                      neatsvpv(tuples,0), neatsvpv(tuples_status,0),
+                      neatsvpv(columns, 0));
+
+    if (is_select) {
+        croak("ora_st_execute_array(): SELECT statement not supported "
+              "for array operation.");
+    }
+
+    if (imp_sth->out_params_av || imp_sth->has_lobs) {
+        croak("ora_st_execute_array(): Output placeholders and LOBs not "
+              "supported for array operation.");
+    }
+
+    /* Check that the `tuples' parameter is an array ref, find the length,
+       and store it in the statement handle for the OCI callback. */
+    if(!SvROK(tuples) || SvTYPE(SvRV(tuples)) != SVt_PVAV) {
+        croak("ora_st_execute_array(): Not an array reference.");
+    }
+    tuples_av = (AV*)SvRV(tuples);
+
+    /* Check the `columns' parameter. */
+    if(SvTRUE(columns)) {
+        if(!SvROK(columns) || SvTYPE(SvRV(columns)) != SVt_PVAV) {
+          croak("ora_st_execute_array(): columns not an array peference.");
+        }
+        columns_av = (AV*)SvRV(columns);
+    } else {
+        columns_av = NULL;
+    }
+
+    /* Check the `tuples_status' parameter. */
+    if(SvTRUE(tuples_status)) {
+        if(!SvROK(tuples_status) || SvTYPE(SvRV(tuples_status)) != SVt_PVAV) {
+          croak("ora_st_execute_array(): tuples_status not an array reference.");
+        }
+        tuples_status_av = (AV*)SvRV(tuples_status);
+        av_fill(tuples_status_av, exe_count - 1);
+        /* Fill in 'unknown' exe count in every element (know not how to get
+           individual execute row counts from OCI). */
+        for(i = 0; (unsigned int) i < exe_count; i++) {
+            av_store(tuples_status_av, i, newSViv((IV)-1));
+        }
+    } else {
+        tuples_status_av = NULL;
+    }
+
+    /* Nothing to do if no tuples. */
+    if(exe_count <= 0)
+      return 0;
+
+    /* Ensure proper OCIBindByName() calls for all placeholders. */
+    if(!ora_st_bind_for_array_exec(sth, imp_sth, tuples_av, exe_count,
+                                   DBIc_NUM_PARAMS(imp_sth), columns_av))
+        return -2;
+
+    /* Store array of bind typles, for use in OCIBindDynamic() callback. */
+    imp_sth->bind_tuples = tuples_av;
+    imp_sth->rowwise = (columns_av == NULL);
+
+    oci_mode = OCI_BATCH_ERRORS;
+    if(autocommit)
+        oci_mode |= OCI_COMMIT_ON_SUCCESS;
+    OCIStmtExecute_log_stat(imp_sth->svchp, imp_sth->stmhp, imp_sth->errhp,
+                            exe_count, 0, 0, 0, oci_mode, exe_status);
+    imp_sth->bind_tuples = NULL;
+
+    if (exe_status != OCI_SUCCESS) {
+ oci_error(sth, imp_sth->errhp, exe_status, ora_sql_error(imp_sth,"OCIStmtExecute"));
+        if(exe_status != OCI_SUCCESS_WITH_INFO)
+            return -2;
+    }
+
+    OCIAttrGet_stmhp_stat(imp_sth, &num_errs, 0, OCI_ATTR_NUM_DML_ERRORS, status);
+    if (debug >= 6)
+ PerlIO_printf(DBILOGFP, "    ora_st_execute_array %d errors in batch.\n",
+                      num_errs);
+    if(num_errs && tuples_status_av) {
+        OCIError *row_errhp, *tmp_errhp;
+        ub4 row_off;
+        SV *err_svs[2];
+        /*AV *err_av;*/
+        sb4 err_code;
+
+        err_svs[0] = newSViv((IV)0);
+        err_svs[1] = newSVpvn("", 0);
+        OCIHandleAlloc_ok(imp_sth->envhp, &row_errhp, OCI_HTYPE_ERROR, status);
+        OCIHandleAlloc_ok(imp_sth->envhp, &tmp_errhp, OCI_HTYPE_ERROR, status);
+        for(i = 0; (unsigned int) i < num_errs; i++) {
+            OCIParamGet_log_stat(imp_sth->errhp, OCI_HTYPE_ERROR,
+                                 tmp_errhp, (dvoid *)&row_errhp,
+                                 (ub4)i, status);
+            OCIAttrGet_log_stat(row_errhp, OCI_HTYPE_ERROR, &row_off, 0,
+                                OCI_ATTR_DML_ROW_OFFSET, imp_sth->errhp, status);
+            if (debug >= 6)
+                PerlIO_printf(DBILOGFP, "    ora_st_execute_array error in row %d.\n",
+                              row_off);
+            sv_setpv(err_svs[1], "");
+            err_code = oci_error_get(row_errhp, exe_status, NULL, err_svs[1], debug);
+            sv_setiv(err_svs[0], (IV)err_code);
+            av_store(tuples_status_av, row_off,
+                     newRV_noinc((SV *)(av_make(2, err_svs))));
+        }
+        OCIHandleFree_log_stat(tmp_errhp, OCI_HTYPE_ERROR,  status);
+        OCIHandleFree_log_stat(row_errhp, OCI_HTYPE_ERROR,  status);
+
+        /* Do a commit here if autocommit is set, since Oracle
+           doesn't do that for us when some rows are in error. */
+        if(autocommit) {
+            OCITransCommit_log_stat(imp_sth->svchp, imp_sth->errhp,
+                                    OCI_DEFAULT, status);
+            if (status != OCI_SUCCESS) {
+                oci_error(sth, imp_sth->errhp, status, "OCITransCommit");
+                return -2;
+            }
+        }
+    }
+
+    if(num_errs) {
+        return -2;
+    } else {
+        ub4 row_count = 0;
+ OCIAttrGet_stmhp_stat(imp_sth, &row_count, 0, OCI_ATTR_ROW_COUNT, status);
+        return row_count;
+    }
+}
+
 
 
 
@@ -1755,7 +2158,7 @@ dbd_st_blob_read(SV *sth, imp_sth_t *imp_sth, int field, long offset, long len,
 
 #ifdef UTF8_SUPPORT
     if (ftype == 112 && CS_IS_UTF8(ncharsetid) ) {
-      return ora_blob_read_mb_piece(sth, imp_sth, fbh, bufsv, 
+      return ora_blob_read_mb_piece(sth, imp_sth, fbh, bufsv,
 				    offset, len, destoffset);
     }
 #endif /* UTF8_SUPPORT */
@@ -1856,6 +2259,7 @@ ora_free_phs_contents(phs_t *phs)
 {
     if (phs->desc_h)
 	OCIDescriptorFree_log(phs->desc_h, phs->desc_t);
+
     sv_free(phs->ora_field);
     sv_free(phs->sv);
 }
@@ -1941,8 +2345,14 @@ dbd_st_destroy(SV *sth, imp_sth_t *imp_sth)
 	hv_iterinit(hv);
 	while( (sv = hv_iternextsv(hv, &key, &retlen)) != NULL ) {
 	    if (sv != &sv_undef) {
-		phs_t *phs = (phs_t*)(void*)SvPVX(sv);
-		ora_free_phs_contents(phs);
+		  phs_t *phs = (phs_t*)(void*)SvPVX(sv);
+
+
+	      if (phs->desc_h && phs->desc_t == OCI_DTYPE_LOB)
+	        ora_free_templob(sth, imp_sth, (OCILobLocator*)phs->desc_h);
+
+
+	      ora_free_phs_contents(phs);
 	    }
 	}
 	sv_free((SV*)imp_sth->all_params_hv);
@@ -1986,7 +2396,7 @@ dbd_st_FETCH_attrib(SV *sth, imp_sth_t *imp_sth, SV *keysv)
     /* int oraperl = DBIc_COMPAT(imp_sth); */
 
     if (kl==13 && strEQ(key, "NUM_OF_PARAMS"))	/* handled by DBI */
-	return Nullsv;	
+	return Nullsv;
 
     if (!imp_sth->done_desc && !dbd_describe(sth, imp_sth)) {
 	STRLEN lna;
